/*
 * some tools to initialize the frame buffer. Inspired from fbtools by
 * Gerd Knorr.
 */
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <signal.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <time.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <linux/kd.h>
#include <linux/vt.h>
#include <linux/fb.h>
#include <termios.h>
#include "fbtools.h"

#define DEFAULT_DEVICE  "/dev/fb0"

#define TTY_SUPPORT 1

struct fb_fix_screeninfo   fb_fix;
struct fb_var_screeninfo   fb_var;
unsigned char             *fb_mem;

static int                       fb;

static struct fb_var_screeninfo  fb_ovar;
static unsigned short            ored[256], ogreen[256], oblue[256];
static struct fb_cmap            ocmap = { 0, 256, ored, ogreen, oblue };

/* tty */
int tty=-1;
struct termios term_settings;
static struct vt_mode            vt_mode;
static int                       kd_mode;
static struct vt_mode            vt_omode;
int fb_switch_state = FB_ACTIVE;

#if TTY_SUPPORT
/* -------------------------------------------------------------------- */
/* console switching                                                    */


static void fb_switch_signal(int signal)
{
    if (signal == SIGUSR1) {
	/* release */
	fb_switch_state = FB_REL_REQ;
    }
    if (signal == SIGUSR2) {
	/* acquisition */
	fb_switch_state = FB_ACQ_REQ;
    }
}

void fb_switch_release()
{
    ioctl(tty, VT_RELDISP, 1);
    fb_switch_state = FB_INACTIVE;
}

void fb_switch_acquire()
{
    ioctl(tty, VT_RELDISP, VT_ACKACQ);
    fb_switch_state = FB_ACTIVE;
}

int fb_switch_init()
{
    struct sigaction act,old;

    act.sa_handler  = fb_switch_signal;
    act.sa_flags    = 0;
    act.sa_restorer = NULL;
    sigemptyset(&act.sa_mask);
    sigaction(SIGUSR1,&act,&old);
    sigaction(SIGUSR2,&act,&old);
    
    if (-1 == ioctl(tty,VT_GETMODE, &vt_mode)) {
	perror("ioctl VT_GETMODE");
	exit(1);
    }
    vt_mode.mode   = VT_PROCESS;
    vt_mode.waitv  = 0;
    vt_mode.relsig = SIGUSR1;
    vt_mode.acqsig = SIGUSR2;
    
    if (-1 == ioctl(tty,VT_SETMODE, &vt_mode)) {
	perror("ioctl VT_SETMODE");
	exit(1);
    }
    return 0;
}

/* tty mode */

void tty_init(void)
{
    struct termios settings;
    int err;

#if 0
    {
        int tmp;
        /* non blocking mode on the tty */
        tmp=fcntl(tty,F_GETFL);
        if (tmp == -1) {
            perror("fnctl");
            exit(1);
        }
        fcntl(tty,F_SETFL,tmp | O_NONBLOCK);
    }
#endif
    
    /* raw mode */
    err=tcgetattr (tty, &settings);
    if (err < 0) {
        perror("tcgetattr");
        exit(1);
    }
    term_settings=settings;

    settings.c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP
                          |INLCR|IGNCR|ICRNL|IXON);
    settings.c_oflag |= OPOST;
#if 0
    settings.c_lflag &= ~(ECHO|ECHONL|ICANON|IEXTEN|ISIG); 
#else
    /* we accept Ctrl C */
    settings.c_lflag &= ~(ECHO|ECHONL|ICANON|IEXTEN); 
#endif
    settings.c_cflag &= ~(CSIZE|PARENB);
    settings.c_cflag |= CS8;
    
    err=tcsetattr (tty, TCSANOW, &settings);
    if (err < 0) {
        perror("tcsetattr");
        exit(1);
    }
}

void tty_cleanup(void)
{
    int err;

    /* put back the terminal settings */
    err=tcsetattr (tty, TCSANOW, &term_settings);
    if (err < 0) {
        perror("tcsetattr");
        exit(1);
    }
}

#endif

/* frame buffer init */

void fb_cleanup(void)
{
    /* restore console */
    if (-1 == ioctl(fb,FBIOPUT_VSCREENINFO,&fb_ovar))
	perror("ioctl FBIOPUT_VSCREENINFO");
    if (fb_ovar.bits_per_pixel == 8)
	if (-1 == ioctl(fb,FBIOPUTCMAP,&ocmap))
	    perror("ioctl FBIOPUTCMAP");
    close(fb);

#if TTY_SUPPORT
    if (-1 == ioctl(tty,KDSETMODE, kd_mode))
        perror("ioctl KDSETMODE");
    if (-1 == ioctl(tty,VT_SETMODE, &vt_omode))
        perror("ioctl VT_SETMODE");
#endif
}

int fb_init(char *device,char *mode)
{
    /* get device handle */
    if (NULL == device)
	if (NULL == (device = getenv("FRAMEBUFFER")))
	    device = (char *) DEFAULT_DEVICE;

    /* get current settings (which we have to restore) */
    if (-1 == (fb = open(device,O_RDWR))) {
	fprintf(stderr,"open %s: %s\n",device,strerror(errno));
	goto err;
    }
    if (-1 == ioctl(fb,FBIOGET_VSCREENINFO,&fb_ovar)) {
	perror("ioctl FBIOGET_VSCREENINFO");
	exit(1);
    }

#if TTY_SUPPORT 
    if (isatty(0)) {
	tty = 0;
    } else if (-1 == (tty = open("/dev/tty0",O_RDWR))) {
	fprintf(stderr,"open %s: %s\n","/dev/tty0",strerror(errno));
	exit(1);
    }

    if (-1 == ioctl(tty,KDGETMODE, &kd_mode)) {
	perror("ioctl KDGETMODE");
	exit(1);
    }
    if (-1 == ioctl(tty,VT_GETMODE, &vt_omode)) {
	perror("ioctl VT_GETMODE");
	exit(1);
    }
#endif

    /* mode switch : NONE yet */
    fb_var=fb_ovar;

    /* checks & initialisation */
    if (-1 == ioctl(fb,FBIOGET_FSCREENINFO,&fb_fix)) {
	perror("ioctl FBIOGET_FSCREENINFO");
	exit(1);
    }
    if (fb_fix.type != FB_TYPE_PACKED_PIXELS) {
	fprintf(stderr,"can handle only packed pixel frame buffers\n");
	goto err;
    }

    /* mmap the frame buffer */
    fb_mem = mmap(NULL,fb_fix.smem_len,PROT_WRITE,MAP_SHARED,fb,0);
    if (-1 == (int)fb_mem) {
	perror("mmap");
	goto err;
    }

    /* move viewport to upper left corner */
    if (fb_var.xoffset != 0 || fb_var.yoffset != 0) {
	fb_var.xoffset = 0;
	fb_var.yoffset = 0;
	if (-1 == ioctl(fb,FBIOPAN_DISPLAY,&fb_var)) {
	    perror("ioctl FBIOPAN_DISPLAY");
	    goto err;
	}
    }
#if TTY_SUPPORT
    if (-1 == ioctl(tty,KDSETMODE, KD_GRAPHICS)) {
	perror("ioctl KDSETMODE");
	goto err;
    }
#endif

    /* cls */
    memset(fb_mem,0,fb_fix.smem_len);
    return fb;

 err:
    fb_cleanup();
    return -1;
}

